#
# KNOWN ISSUES
#
# - Selects second match on down key press (should select the first)
#
# - Up key press in entry does not select bottom hit
#
# - In window mode Up key press from list_views top hit does not focus entry,
#   but it does in applet mode...
#

from os.path import *
from gettext import gettext as _

import cgi
import sys

import gtk

import gnome, gobject, gconf
import gnome.ui, gnomeapplet
import pango

import deskbar
from deskbar.ui import EntryHistoryManager
from deskbar.ui.DeskbarUI import DeskbarUI
from deskbar.ui.cuemiac.CuemiacAlignedWindow import CuemiacAlignedWindow
from deskbar.ui.cuemiac.CuemiacHistory import CuemiacHistoryPopup
from deskbar.DeskbarHistory import get_deskbar_history
from deskbar.ui.EntryHistoryManager import EntryHistoryManager

			
# The sort function ids
SORT_BY_CATEGORY = 1

class FlatModel (gtk.ListStore):

	# Column name
	MATCHES = 0
	ACTIONS = 1
	HANDLER_PRIO_COL = 2
	MATCH_PRIO_COL = 3
	
	def __init__ (self):
		gtk.ListStore.__init__ (self, gobject.TYPE_PYOBJECT, gobject.TYPE_STRING, gobject.TYPE_INT, gobject.TYPE_INT)
		self.append_method = gtk.ListStore.append # Alternatively gtk.TreeStore.prepend for bottom panel layout
		self.set_sort_func(SORT_BY_CATEGORY, self.__on_sort_matches)
	
	def set_sort_order(self, order):
		self.set_sort_column_id(SORT_BY_CATEGORY, order)
	
	def __on_sort_matches(self, treemodel, iter1, iter2):
		# First compare global handler priority
		diff = treemodel[iter1][self.HANDLER_PRIO_COL] - treemodel[iter2][self.HANDLER_PRIO_COL]
		if diff != 0:
			return diff
		
		# We have the same global priority, use relative priority between matches
		diff = treemodel[iter1][self.MATCH_PRIO_COL] - treemodel[iter2][self.MATCH_PRIO_COL]
		if diff != 0:
			return diff
		
		# Finally use the Action to sort alphabetically
		a = treemodel[iter1][self.ACTIONS]
		b = treemodel[iter2][self.ACTIONS]
		if a != None:
			a = a.strip().lower()
		if b != None:
			b = b.strip().lower()
			
		if a < b:
			return 1
		elif a > b:
			return -1
		else:
			return 0
		
	def append (self, match):
		"""
		Automagically append a match or list of matches 
		to correct category(s), or create a new one(s) if needed.
		"""

		if type (match) == list:
			for hit in match:
				self.__append (hit)
		else:
			self.__append (match)
			
	def __append (self, match):
		"""match is a single tuple (qstring,match_obj)"""
		qstring, match_obj = match
		# Pass unescaped query to the matches
		verbs = {"text" : qstring}
		verbs.update(match_obj.get_name(qstring))
		# Escape the query now for display
		verbs["text"] = cgi.escape(verbs["text"])
		
		handler_priority = match_obj.get_handler().get_priority()
		if hasattr(match_obj, '_history_priority'):
			handler_priority = match._history_priority
		
		self.append_method (self, [match, match_obj.get_verb () % verbs, handler_priority, match_obj.get_priority ()])	
		
	def paths_equal (self, path1, path2):
		"""Returns true if the two paths point to the same cell."""
		if path1 == None or path2 == None:
			return False
			
		return ( self.get_string_from_iter (self.get_iter(path1)) == self.get_string_from_iter (self.get_iter(path2)) )


class FlatListView (gtk.TreeView):
	"""

	"""
	
	activation_keys = [65293] # Enter  - Space makes a mess when users type in queries with spaces
	
	__gsignals__ = {
		"match-selected" : (gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, [gobject.TYPE_PYOBJECT]),
	}
	
	def __init__ (self, model):
		gtk.TreeView.__init__ (self, model)
				
		icon = gtk.CellRendererPixbuf ()
		hit_title = gtk.CellRendererText ()
		hit_title.set_property ("ellipsize", pango.ELLIPSIZE_END)
		hit_title.set_property ("width-chars", 50) #FIXME: Pick width according to screen size
		hits = gtk.TreeViewColumn ("Hits")
		hits.pack_start (icon)
		hits.pack_start (hit_title)
		hits.set_cell_data_func(hit_title, self.__get_match_title_for_cell)			
		hits.set_cell_data_func(icon, self.__get_match_icon_for_cell)
		self.append_column (hits)
		
		self.set_property ("headers-visible", False)
		self.connect ("row-activated", self.__on_activated)
		#self.connect ("key-press-event", self.__on_key_press)
				
		self.set_enable_search (False)
		
		self.qstring = ""

	def clear (self):
		self.model.clear ()
	
	def set_query_string (self, qstring):
		self.qstring = qstring
	
	def last_visible_path (self):
		"""Returns the path to the last visible cell."""
		model = self.get_model ()
		iter = model.get_iter_first ()
		next = model.iter_next (iter)
		
		# Find last category node (they are always visible)
		while next:
			iter = next
			next = model.iter_next (iter)
		
		# If this category is not expanded return
		if not self.row_expanded (model.get_path (iter)):
			return model.get_path (iter)
		
		return model.get_path(iter)
	
	def __get_match_icon_for_cell (self, column, cell, model, iter, data=None):
	
		qstring, match_obj = model[iter][model.MATCHES]
		cell.set_property ("pixbuf", match_obj.get_icon())

		
	def __get_match_title_for_cell (self, column, cell, model, iter, data=None):
	
		match = model[iter][model.MATCHES]
		cell.set_property ("markup", model[iter][model.ACTIONS])

	def __on_activated (self, treeview, path, colum):
		model, iter = self.get_selection().get_selected()
		if iter is None:
			return
		match = model[iter][model.MATCHES]
		self.emit ("match-selected", match)
				
	def __on_key_press (self, widget, event):
		# FIXME: In the future we should check for ctrl being pressed to create shortcuts
		return False
		

class FlatListUI (DeskbarUI):
	
	navigation_keys = [65364, 65362, 65293] # Down, Up, Enter
	
	def __init__ (self, applet, prefs):
		DeskbarUI.__init__ (self, applet, prefs)
		
		self.default_entry_pixbuf = deskbar.Utils.load_icon("deskbar-applet-small.png", width=-1)
		
		self.icon_entry = deskbar.iconentry.IconEntry ()
		self.icon_event_box = gtk.EventBox ()
		self.popup = CuemiacAlignedWindow (self.icon_entry, applet)
		self.entry = self.icon_entry.get_entry ()
		self.entry_icon = gtk.Image ()
		self.history = get_deskbar_history ()
		self.history_popup = CuemiacHistoryPopup (self.history, self.entry_icon, applet)
		self.model = FlatModel ()
		self.list_view = FlatListView (self.model)
		self.scroll_win = gtk.ScrolledWindow ()
		self.scroll_win.set_policy (gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)	
		
		self.set_layout_by_orientation (applet.get_orient())
			
		self.popup.add (self.scroll_win)
		self.scroll_win.add(self.list_view)
		
		# Set up the event box for the entry icon
		self.icon_event_box.set_property('visible-window', False)
		self.icon_event_box.add(self.entry_icon)
		self.entry_icon.set_property('pixbuf', self.default_entry_pixbuf)
		self.icon_entry.pack_widget (self.icon_event_box, True)
		
		self.popup.set_border_width (1)
		self.history_popup.set_border_width (1)
		
		self.scroll_win.connect ("size-request", lambda win, event: self.adjust_popup_size())
		on_entry_changed_id = self.entry.connect ("changed", self.on_entry_changed)
		
		# Connect first the history handler then the regular key handler
		self.history_entry_manager = EntryHistoryManager(self.entry, on_entry_changed_id)
		self.history_entry_manager.connect('history-set', self.on_history_set)
		self.icon_event_box.connect ("button-press-event", self.on_icon_button_press)
		
		self.entry.connect ("key-press-event", self.on_entry_key_press)
		self.entry.connect_after ("changed", lambda entry : self.update_entry_icon())
		self.entry.connect ("activate", self.on_entry_activate)
		self.entry.connect ("button-press-event", self.on_entry_button_press)
		self.list_view.connect ("key-press-event", self.on_list_view_key_press)
		self.list_view.connect ("match-selected", self.on_match_selected)
		self.list_view.connect_after ("cursor-changed", lambda treeview : self.update_entry_icon())
		self.history_popup.connect ("match-selected", self.on_match_selected, True)
		self.history_popup.connect ("key-press-event", self.on_history_key_press)		
		
		# Setup a bunch of window hiding conditions
		self.list_view.connect ("focus-out-event", lambda widget, event : self.hide_window(self.popup))
		#self.entry.connect ("focus-out-event", self.hide_windows)
		self.history_popup.connect ("focus-out-event", lambda widget, event : self.hide_window (self.history_popup))
		
		self.screen_height = self.popup.get_screen().get_height ()
		self.screen_width = self.popup.get_screen().get_width ()
		self.max_window_height = int (0.8 * self.screen_height)
		self.max_window_width = int (0.6 * self.screen_width)

		# Show child widgets of the various windows, so we only need to show() not show_all()
		self.scroll_win.show_all ()
		self.icon_entry.show_all ()
		self.history_popup.get_child().show ()
		
		self.set_sensitive(False)
		try:
			self.applet.set_background_widget(self.deskbar_button)
		except Exception, msg:
			pass
		
		self.invalid = True
		
		self.applet.set_applet_flags(gnomeapplet.EXPAND_MINOR)
		#self.applet.set_flags(gtk.CAN_FOCUS)
	
	def hide_window (self, window):
		self.list_view.get_selection().unselect_all ()
		self.history_popup.list_view.get_selection().unselect_all ()
		window.hide ()
		self.update_entry_icon ()
	
	def update_entry_icon (self, icon=None):
		if icon == None:
			icon = self.default_entry_pixbuf
			path, column = self.list_view.get_cursor ()
		
			if path != None:
				qstring, match_obj = self.model[self.model.get_iter(path)][self.model.MATCHES]
				icon = match_obj.get_icon()
			else:
				first = self.model.get_iter_first()
				if first:
					qstring, match_obj = self.model[first][self.model.MATCHES]
					icon = match_obj.get_icon()
				
		self.entry_icon.set_property('pixbuf', icon)
		
	def on_match_selected (self, list_iew, match, historic=False):
		self.emit ("match-selected", match[0], match[1])
		self.history_popup.hide ()
		self.popup.hide ()
		self.update_entry_icon ()
	
	def show_popup (self, time=None):
		self.adjust_popup_size ()
		self.popup.update_position ()
		self.update_entry_icon ()
		if time:
			self.popup.present_with_time (time)
		else:
			self.popup.present ()
	
	def on_icon_button_press(self, widget, event):
		if not self.icon_event_box.get_property('sensitive'):
			return False
			
		if event.button == 3:
			self.applet.emit("button-press-event", event)
			return True
		elif event.button == 1:
			self.history_popup.update_position ()
			self.history_popup.present ()
			return True
		
		return False
	
	def receive_focus (self, time):
		# Left-Mouse-Button should focus the GtkEntry widget (for Fitt's Law
		# - so that a click on applet border on edge of screen activates the
		# most important widget).
		try:
			# GNOME 2.12
			self.applet.request_focus(long(time))
		except AttributeError:
			pass
			
		self.entry.select_region(0, -1)
		self.entry.grab_focus()
	
	def get_view (self):
		return self.icon_entry
		
	def set_sensitive (self, active):
		self.icon_entry.set_sensitive (active)
		self.icon_event_box.set_sensitive (active)
		
	def on_change_orient (self, applet):
		self.set_layout_by_orientation (applet.get_orient())
	
	def append_matches (self, matches):
		if self.invalid :
			self.invalid = False
			self.model.clear()
		entry_text = self.entry.get_text().strip()
		valid_matches = False
		for qstring, match_obj in matches:
			if qstring == entry_text: # FIXME: Maybe it will suffice to only check the first match
				self.model.append ((qstring,match_obj))
				valid_matches = True
		if valid_matches:
			self.show_popup ()
		else:
			print " *** No valid matches!"
		
	def set_layout_by_orientation (self, orient):
		"""orient should be a gnomeapplet.ORIENT_{UP,DOWN,LEFT,RIGHT}."""

		if orient in [gnomeapplet.ORIENT_LEFT, gnomeapplet.ORIENT_RIGHT, gnomeapplet.ORIENT_DOWN]:
			self.list_view.append_method = gtk.ListStore.append
			self.list_view.get_model().set_sort_order(gtk.SORT_DESCENDING)
		else:
			# We are at a bottom panel. Put entry on bottom, and prepend matches (instead of append).
			self.list_view.append_method = gtk.ListStore.prepend
			self.list_view.get_model().set_sort_order(gtk.SORT_ASCENDING)
		
		# Update how the popups is aligned
		self.popup.alignment = self.applet.get_orient ()
		self.history_popup.alignment = self.applet.get_orient ()
		
		
	def adjust_popup_size (self):
		"""adjust window size to the size of the children"""
		# FIXME: Should we handle width intelligently also?
		w, h = self.list_view.size_request ()
		h = min (h, self.max_window_height) + 4
		w = min (w, self.max_window_width)
		if w > 0 and h > 0:
			self.popup.resize (w, h)
		
	def on_entry_changed (self, entry):
		self.history.reset()
		qstring = self.entry.get_text().strip()
		self.list_view.set_query_string (qstring)
		if qstring == "":
			self.model.clear()
			self.popup.hide ()
			self.emit ("stop-query")
			return
		
		self.invalid = True
		self.show_popup ()
		self.emit ("start-query", qstring, 100)
	
	def on_entry_key_press (self, entry, event):
		
		if event.keyval == gtk.keysyms.Escape:
			# bind Escape to clear the GtkEntry
			self.emit ("stop-query")	
			self.entry.set_text ("")	
			return True
		
		if 	event.state&gtk.gdk.MOD1_MASK != 0:
			# Some Handlers want to know about Alt-keypress
			# combinations, for example.  Here, we notify such
			# Handlers.
			text = entry.get_text().strip()
			if text != "":
				self.emit('stop-query')
				self.emit('keyboard-shortcut', text, event.keyval)
			entry.set_text("")
			
			# Broadcast an escape
			event.state = 0
			event.keyval = gtk.keysyms.Escape
			entry.emit('key-press-event', event)
			return True
			
		if event.keyval == 65362: # Up
			self.show_popup (event.time)
			self.list_view.grab_focus ()
			return True
			
		if event.keyval == 65364: # Down
			self.show_popup (event.time)
			self.list_view.grab_focus ()
			last = self.list_view.get_last_visible_path ()
			self.list_view.set_cursor (last)
			print self.model.iter_to_string (self.model.get_iter (last))
			return True

		return False
		
	def on_history_key_press (self, history, event):
		if event.keyval == gtk.keysyms.Escape:
			self.history_popup.hide ()
			
	def on_entry_activate(self, widget):
		# if we have an active history item, use it
		if self.history_entry_manager.current_history != None:
			text, match = self.history_entry_manager.current_history
			self.on_match_selected(widget, (text, match))
			return
			
		path, column = self.list_view.get_cursor ()
		iter = None
		if path != None:
			iter = self.model.get_iter (path)
			
		if iter is None:
			# No selection, select top element
			iter = self.model.get_iter_first()
			
		if iter is None:
			return
						
		self.on_match_selected(widget, self.model[iter][self.model.MATCHES])


	def on_history_set(self, historymanager, set):
		if set:
			text, match = historymanager.current_history
			self.update_entry_icon (icon=match.get_icon())
		else:
			self.entry.set_text("")
			
	def on_list_view_key_press (self, list_view, event):
		# If this is an ordinary keystroke just let the
		# entry handle it.
		if not event.keyval in self.navigation_keys:
			self.entry.event (event)
			return True
			
		path, column = list_view.get_cursor ()
		model = list_view.get_model ()
		if model.paths_equal (path, model.get_path(model.get_iter_first())):
			if event.keyval == 65362: # Up
				self.list_view.get_selection().unselect_all ()
				self.applet.request_focus(long(event.time))
				self.entry.grab_focus ()
			
		elif model.paths_equal (path, list_view.last_visible_path()):
			if event.keyval == 65364: # Down
				self.list_view.get_selection().unselect_all ()
				self.applet.request_focus(long(event.time))
				self.entry.grab_focus ()

		return False		

	def on_entry_button_press(self, widget, event):
		try:
			# GNOME 2.12
			self.applet.request_focus(long(event.time))
		except AttributeError:
			pass
			
		return False

gobject.type_register (FlatListUI)

if gtk.pygtk_version < (2,8,0):	
	gobject.type_register (FlatListView)
	gobject.type_register (FlatModel)
